import json

from aiohttp import web
from nacl.exceptions import BadSignatureError
from nacl.public import PublicKey
from nacl.signing import VerifyKey
from nacl.utils import random
from v6d0auth import certs
from v6d0auth.appfactory import AppFactory
from v6d0auth.client import has_role

from v6d1tokens.tdb import TDB

__all__ = ('V6D1TokensAppFactory',)


class V6D1TokensAppFactory(AppFactory):
    def __init__(self, tdb: TDB):
        self.tdb = tdb

    def define_routes(self, routes: web.RouteTableDef):
        print(certs.vkey.encode().hex())

        @routes.get('/')
        async def home(_request: web.Request):
            return web.Response(body='v6d1tokens\n')

        async def ws_reg(ws: web.WebSocketResponse):
            nonce = random(16)
            await ws.send_bytes(nonce)
            [token_id, token], hnonce = json.loads(certs.verify(await ws.receive_bytes()))
            assert hnonce == nonce.hex()
            await self.tdb.reg(token_id, token)
            await ws.send_bytes(b'1')
            await ws.close()

        @routes.get('/reg')
        async def reg(request: web.Request):
            ws = web.WebSocketResponse()
            await ws.prepare(request)
            await ws_reg(ws)
            return ws

        def role_matches_token_id(role: str, token_id: str) -> bool:
            return role.startswith(f'token:{token_id}::') or token_id.startswith(f'role:{role}::')

        async def requester_for_request(request: web.Request, requester_cert: bytes, token_id: str) -> VerifyKey:
            role = request.headers.get('v6role')
            if role is None:
                return VerifyKey(certs.averify(requester_cert))
            else:
                requester = VerifyKey(requester_cert)
                assert (await has_role(requester, role))
                assert role_matches_token_id(role, token_id)
                return requester

        @routes.post('/get')
        async def get(request: web.Request):
            try:
                token_id, requester_hcert = json.loads(await request.read())
                requester_cert = bytes.fromhex(requester_hcert)
                requester = await requester_for_request(request, requester_cert, token_id)
                requesterpk: PublicKey = requester.to_curve25519_public_key()
                token_encrypted = await self.tdb.get_encrypted(requesterpk, token_id)
            except BadSignatureError:
                raise web.HTTPUnauthorized
            except (json.JSONDecodeError, AssertionError):
                raise web.HTTPBadRequest
            except KeyError:
                raise web.HTTPNotFound
            else:
                return web.Response(body=token_encrypted)
